import os
import unittest
import numpy as np

from gnes.encoder.audio.vggish import VggishEncoder


class TestVggishEncoder(unittest.TestCase):
    @unittest.skip
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.video_path = os.path.join(self.dirname, 'videos')
        self.video_bytes = [open(os.path.join(self.video_path, _), 'rb').read()
                            for _ in os.listdir(self.video_path)]
        self.audios = [np.random.rand(10, 96, 64),
                       np.random.rand(15, 96, 64),
                       np.random.rand(5, 96, 64)]
        self.vggish_yaml = os.path.join(self.dirname, 'yaml', 'vggish-encoder.yml')

    @unittest.skip
    def test_vggish_encoding(self):
        self.encoder = VggishEncoder.load_yaml(self.vggish_yaml)
        vec = self.encoder.encode(self.audios)
        self.assertEqual(len(vec.shape), 2)
        self.assertEqual(vec.shape[0], len(self.audios))
        self.assertEqual(vec.shape[1], 128)


#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import threading
from concurrent.futures import ThreadPoolExecutor

import grpc

from .. import __version__, __proto_version__
from ..client.base import ZmqClient
from ..helper import set_logger
from ..proto import gnes_pb2_grpc, gnes_pb2, router2str, add_route


class FrontendService:

    def __init__(self, args):
        self.logger = set_logger(self.__class__.__name__, args.verbose)
        self.server = grpc.server(
            ThreadPoolExecutor(max_workers=args.max_concurrency),
            options=[('grpc.max_send_message_length', args.max_message_size),
                     ('grpc.max_receive_message_length', args.max_message_size)])
        self.logger.info('start a frontend with %d workers' % args.max_concurrency)
        gnes_pb2_grpc.add_GnesRPCServicer_to_server(self._Servicer(args), self.server)

        self.bind_address = '{0}:{1}'.format(args.grpc_host, args.grpc_port)
        self.server.add_insecure_port(self.bind_address)

    def __enter__(self):
        self.server.start()
        self.logger.critical('listening at: %s' % self.bind_address)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.server.stop(None)

    class _Servicer(gnes_pb2_grpc.GnesRPCServicer):

        def __init__(self, args):
            self.args = args
            self.logger = set_logger(FrontendService.__name__, args.verbose)
            self.zmq_context = self.ZmqContext(args)
            self.request_id_cnt = 0

        def add_envelope(self, body: 'gnes_pb2.Request', zmq_client: 'ZmqClient'):
            msg = gnes_pb2.Message()
            msg.envelope.client_id = zmq_client.args.identity
            if body.request_id is not None:
                msg.envelope.request_id = body.request_id
            else:
                msg.envelope.request_id = self.request_id_cnt
                self.request_id_cnt += 1
                self.logger.warning('request_id is missing, filled it with an internal counter!')
            msg.envelope.part_id = 1
            msg.envelope.num_part.append(1)
            msg.envelope.timeout = 5000
            msg.envelope.gnes_version = __version__
            msg.envelope.proto_version = __proto_version__
            add_route(msg.envelope, FrontendService.__name__, self.args.identity)
            msg.request.CopyFrom(body)
            return msg

        def remove_envelope(self, m: 'gnes_pb2.Message'):
            resp = m.response
            resp.request_id = m.envelope.request_id
            m.envelope.routes[0].end_time.GetCurrentTime()
            if self.args.route_table:
                self.logger.info('route: %s' % router2str(m))
                route_time = []
                k = m.envelope.routes[0]
                total_duration = self.get_duration(k.start_time, k.end_time)

                sum_duration = 0
                for k in m.envelope.routes[1:]:
                    if k.first_start_time and k.last_end_time:
                        d = self.get_duration(k.first_start_time, k.last_end_time)
                    else:
                        d = self.get_duration(k.start_time, k.end_time)

                    route_time.append((k.service, d))
                    sum_duration += d

                def get_table_str(time_table):
                    return '\n'.join(
                        ['%40s\t%3.3fs\t%3d%%' % (k[0], k[1], k[1] / total_duration * 100) for k in
                         sorted(time_table, key=lambda x: x[1], reverse=True)])

                summary = [('system', total_duration - sum_duration),
                           ('total', total_duration),
                           ('job', sum_duration)]

                route_table = ('\n%s\n' % ('-' * 80)).join(
                    ['%40s\t%-6s\t%3s' % ('Breakdown', 'Time', 'Percent'), get_table_str(route_time),
                     get_table_str(summary)])
                self.logger.info('route table: \n%s' % route_table)

            return resp

        @staticmethod
        def get_duration(start_time, end_time):
            d_s = end_time.seconds - start_time.seconds
            d_n = end_time.nanos - start_time.nanos
            if d_s < 0 and d_n > 0:
                d_s = max(d_s + 1, 0)
                d_n = max(d_n - 1e9, 0)
            elif d_s > 0 and d_n < 0:
                d_s = max(d_s - 1, 0)
                d_n = max(d_n + 1e9, 0)
            return max(d_s + d_n / 1e9, 0)

        def Call(self, request, context):
            with self.zmq_context as zmq_client:
                zmq_client.send_message(self.add_envelope(request, zmq_client), self.args.timeout)
                return self.remove_envelope(zmq_client.recv_message(self.args.timeout))

        def Train(self, request, context):
            return self.Call(request, context)

        def Index(self, request, context):
            return self.Call(request, context)

        def Search(self, request, context):
            return self.Call(request, context)

        def StreamCall(self, request_iterator, context):
            with self.zmq_context as zmq_client:
                # network traffic control
                num_request = 0
                max_outstanding = 500

                for request in request_iterator:
                    timeout = 25
                    if self.args.timeout > 0:
                        timeout = min(0.5 * self.args.timeout, 50)

                    while num_request > 10:
                        try:
                            msg = zmq_client.recv_message(timeout)
                            yield self.remove_envelope(msg)
                            num_request -= 1
                        except TimeoutError:
                            if num_request > max_outstanding:
                                self.logger.warning("the network traffic exceed max outstanding (%d > %d)" % (num_request, max_outstanding))
                                continue
                            break
                    zmq_client.send_message(self.add_envelope(request, zmq_client), -1)
                    num_request += 1

                for _ in range(num_request):
                    msg = zmq_client.recv_message(self.args.timeout)
                    yield self.remove_envelope(msg)

        class ZmqContext:
            """The zmq context class."""

            def __init__(self, args):
                self.args = args
                self.tlocal = threading.local()
                self.tlocal.client = None

            def __enter__(self):
                """Enter the context."""
                client = ZmqClient(self.args)
                self.tlocal.client = client
                return client

            def __exit__(self, exc_type, exc_value, exc_traceback):
                """Exit the context."""
                self.tlocal.client.close()
                self.tlocal.client = None



import os
import unittest
from shutil import rmtree

import grpc

from gnes.cli.parser import set_frontend_parser, set_preprocessor_parser, set_indexer_parser
from gnes.indexer.base import BaseIndexer
from gnes.indexer.doc.filesys import DirectoryIndexer
from gnes.preprocessor.base import BasePreprocessor
from gnes.proto import gnes_pb2, gnes_pb2_grpc, RequestGenerator
from gnes.service.base import SocketType, ServiceManager
from gnes.service.frontend import FrontendService
from gnes.service.indexer import IndexerService
from gnes.service.preprocessor import PreprocessorService


class TestDictIndexer(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)

        self.video_path = os.path.join(self.dirname, 'videos')
        self.video_bytes = [open(os.path.join(self.video_path, _), 'rb').read()
                            for _ in os.listdir(self.video_path)]

        self.pipeline_name = 'pipe-gif'
        self.pipeline_yml_path = os.path.join(self.dirname, 'yaml/%s.yml' % self.pipeline_name)
        self.data_path = './test_chunkleveldb'
        self.dump_path = os.path.join(self.dirname, 'indexer.bin')

        self.init_db()

    def test_pymode(self):
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')
        args = set_frontend_parser().parse_args([])

        p_args = set_preprocessor_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', '5531',
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_BIND),
            '--yaml_path', 'SentSplitPreprocessor'
        ])

        e_args = set_indexer_parser().parse_args([
            '--port_in', str(p_args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', '!DictIndexer {gnes_config: {name: dummy_dict_indexer}}',
        ])

        with ServiceManager(IndexerService, e_args), \
             ServiceManager(PreprocessorService, p_args), \
             FrontendService(args), \
             grpc.insecure_channel('%s:%s' % (args.grpc_host, args.grpc_port),
                                   options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                                            ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            all_bytes = []
            with open(os.path.join(self.dirname, '26-doc-chinese.txt'), 'r', encoding='utf8') as fp:
                for v in fp:
                    if v.strip():
                        all_bytes.append(v.encode())
            for r in stub.StreamCall(RequestGenerator.index(all_bytes)):
                print(r)

        bi = BaseIndexer.load('dummy_dict_indexer.bin')
        self.assertEqual(bi.num_docs, 26)
        print(bi.query([0]))

    def tearDown(self):
        if os.path.exists(self.data_path):
            rmtree(self.data_path)
        if os.path.exists('dummy_dict_indexer.bin'):
            os.remove('dummy_dict_indexer.bin')

    def init_db(self):
        self.db = DirectoryIndexer(self.data_path)

        self.d = gnes_pb2.Document()
        self.d.doc_id = 0
        self.d.raw_bytes = self.video_bytes[0]

        preprocess = BasePreprocessor.load_yaml(self.pipeline_yml_path)
        preprocess.apply(self.d)

        self.db.add(list(range(len(self.video_bytes))), [self.d])
        self.assertEqual(self.db.num_docs, len(self.video_bytes))

    def test_add_docs(self):
        # self.init_db()
        self.assertTrue(os.path.exists(os.path.join(self.data_path, str(self.d.doc_id))))
        self.assertEqual(len(self.d.chunks), len(os.listdir(os.path.join(self.data_path, str(self.d.doc_id)))) - 1)

    def test_query_docs(self):
        # self.init_db()

        query_list = [0, 1, 2]
        res = self.db.query(query_list)
        num_non_empty = sum(1 for d in res if d)
        self.assertEqual(num_non_empty, 1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
import numpy as np

from typing import List

from .ffmpeg import compile_args, probe
from .helper import _check_input, run_command, run_command_async


def scale_video(input_fn: str = 'pipe:',
                output_fn: str = 'pipe:',
                input_data: bytes = None,
                start_time: float = None,
                end_time: float = None,
                scale: str = None,
                frame_rate: int = 15,
                crf: int = 16,
                vcodec: str = 'libx264',
                format: str = 'mp4',
                pix_fmt: str = 'yuv420p',
                **kwargs):

    _check_input(input_fn, input_data)

    capture_stdout = (output_fn == 'pipe:')

    input_kwargs = {}
    if start_time is not None:
        input_kwargs['ss'] = start_time
    else:
        start_time = 0.
    if end_time is not None:
        input_kwargs['t'] = end_time - start_time

    out_kwargs = {
        'vcodec': vcodec,
        'pix_fmt': pix_fmt,
        'crf': crf,
        'framerate': frame_rate,
        'acodec': 'aac',
        'strict': 'experimental',    # AAC audio encoder is experimental
    }

    if scale:
        out_kwargs['s'] = scale

    if capture_stdout:
        out_kwargs['format'] = format
        # an empty moov means it doesn't need to seek and thus works with a pipe.
        out_kwargs['movflags'] = 'frag_keyframe+empty_moov'

    cmd_args = compile_args(
        input_fn=input_fn,
        output_fn=output_fn,
        input_options=input_kwargs,
        output_options=out_kwargs,
        overwrite_output=True)
    stdout, _ = run_command(
        cmd_args, input=input_data, pipe_stdout=True, pipe_stderr=True)

    if capture_stdout:
        return stdout
    return None


def encode_video(images: 'np.ndarray',
                 pix_fmt: str = 'rgb24',
                 frame_rate: int = 15,
                 output_fn: str = 'pipe:',
                 vcodec: str = 'libx264',
                 format: str = 'mp4',
                 **kwargs):
    packet_size = 4096

    n = len(images)
    height, width, channels = images[0].shape

    capture_stdout = (output_fn == 'pipe:')

    input_kwargs = {
        'format': 'rawvideo',
        'pix_fmt': pix_fmt,
        'framerate': frame_rate,
        's': '{}x{}'.format(width, height),
    }

    output_kwargs = {
        'vcodec': vcodec,
        'r': frame_rate,
        'pix_fmt': 'yuv420p',
        'format': format,
        'movflags': 'frag_keyframe+empty_moov',
    }

    cmd_args = compile_args(
        input_fn='pipe:',
        output_fn=output_fn,
        input_options=input_kwargs,
        output_options=output_kwargs)

    with run_command_async(
            cmd_args,
            pipe_stdin=True,
            pipe_stdout=capture_stdout,
            pipe_stderr=True) as proc:

        input_stream = io.BytesIO(b'')
        for frame in images:
            input_stream.write(frame.astype(np.uint8).tobytes())

        output, err = proc.communicate(input_stream.getvalue())

        if proc.returncode:
            err = '\n'.join([' '.join(cmd_args), err.decode('utf8')])
            raise IOError(err)

        return output


def capture_frames(input_fn: str = 'pipe:',
                   input_data: bytes = None,
                   pix_fmt: str = 'rgb24',
                   fps: int = -1,
                   scale: str = None,
                   start_time: float = None,
                   end_time: float = None,
                   vframes: int = -1,
                   **kwargs) -> List['np.ndarray']:
    _check_input(input_fn, input_data)

    import tempfile

    with tempfile.NamedTemporaryFile() as f:
        if input_data:
            f.write(input_data)
            f.flush()
            input_fn = f.name

        video_meta = probe(input_fn)
        width = video_meta['width']
        height = video_meta['height']

        if scale is not None:
            _width, _height = map(int, scale.split(':'))
            if _width * _height < 0:
                if _width > 0:
                    ratio = _width / width
                    height = int(ratio * height)
                    if _height == -2:
                        height += height % 2
                    width = _width
                else:
                    ratio = _height / height
                    width = int(ratio * width)
                    if _width == -2:
                        width += width % 2

                    height = _height

                scale = '%d:%d' % (width, height)
            else:
                width = _width
                height = _height

        input_kwargs = {
            'err_detect': 'aggressive',
            'fflags': 'discardcorrupt'    # discard corrupted frames
        }
        if start_time is not None:
            input_kwargs['ss'] = str(start_time)
        else:
            start_time = 0.
        if end_time is not None:
            input_kwargs['t'] = str(end_time - start_time)

        video_filters = []
        if fps:
            video_filters += ['fps=%d' % fps]
        if scale:
            video_filters += ['scale=%s' % scale]

        output_kwargs = {
            'format': 'image2pipe',
            'pix_fmt': pix_fmt,
            'vcodec': 'rawvideo',
            'movflags': 'faststart',
        }
        if vframes > 0:
            output_kwargs['vframes'] = vframes

        cmd_args = compile_args(
            input_fn=input_fn,
            input_options=input_kwargs,
            video_filters=video_filters,
            output_options=output_kwargs)
        out, _ = run_command(cmd_args, pipe_stdout=True, pipe_stderr=True)

        depth = 3
        if pix_fmt == 'rgba':
            depth = 4

        frames = np.frombuffer(out,
                               np.uint8).reshape([-1, height, width, depth])
        return frames


# def read_frame_as_jpg(in_filename, frame_num):
#     out, err = (
#         ffmpeg
#         .input(in_filename)
#         .filter_('select', 'gte(n,{})'.format(frame_num))
#         .output('pipe:', vframes=1, format='image2', vcodec='mjpeg')
#         .run(capture_stdout=True)
#     )
#     return out



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import numpy as np

from .base import BaseService as BS, MessageHandler, ServiceError
from ..proto import gnes_pb2, blob2array


class IndexerService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..indexer.base import BaseIndexer
        self._model = self.load_model(BaseIndexer)

    @handler.register(gnes_pb2.Request.IndexRequest)
    def _handler_index(self, msg: 'gnes_pb2.Message'):
        from ..indexer.base import BaseChunkIndexer, BaseDocIndexer
        if isinstance(self._model, BaseChunkIndexer):
            self._handler_chunk_index(msg)
        elif isinstance(self._model, BaseDocIndexer):
            self._handler_doc_index(msg)
        else:
            raise ServiceError(
                'unsupported indexer, dont know how to use %s to handle this message' % self._model.__bases__)
        msg.response.index.status = gnes_pb2.Response.SUCCESS
        self.is_model_changed.set()

    def _handler_chunk_index(self, msg: 'gnes_pb2.Message'):
        vecs, doc_ids, offsets, weights = [], [], [], []

        for d in msg.request.index.docs:
            if not d.chunks:
                self.logger.warning('document (doc_id=%s) contains no chunks!' % d.doc_id)
                continue

            vecs += [blob2array(c.embedding) for c in d.chunks]
            doc_ids += [d.doc_id] * len(d.chunks)
            offsets += [c.offset for c in d.chunks]
            weights += [c.weight for c in d.chunks]

        if vecs:
            self._model.add(list(zip(doc_ids, offsets)), np.stack(vecs), weights)
        else:
            self.logger.warning('chunks contain no embedded vectors, %the indexer will do nothing')

    def _handler_doc_index(self, msg: 'gnes_pb2.Message'):
        self._model.add([d.doc_id for d in msg.request.index.docs],
                        [d for d in msg.request.index.docs],
                        [d.weight for d in msg.request.index.docs])

    def _put_result_into_message(self, results, msg: 'gnes_pb2.Message'):
        msg.response.search.ClearField('topk_results')
        msg.response.search.topk_results.extend(results)
        msg.response.search.top_k = len(results)
        msg.response.search.is_big_score_similar = self._model.is_big_score_similar

    @handler.register(gnes_pb2.Request.QueryRequest)
    def _handler_chunk_search(self, msg: 'gnes_pb2.Message'):
        from ..indexer.base import BaseChunkIndexer
        if not isinstance(self._model, BaseChunkIndexer):
            raise ServiceError(
                'unsupported indexer, dont know how to use %s to handle this message' % self._model.__bases__)

        results = []
        if not msg.request.search.query.chunks:
            self.logger.warning('query contains no chunks!')
        else:
            results = self._model.query_and_score(msg.request.search.query.chunks, top_k=msg.request.search.top_k)

        self._put_result_into_message(results, msg)

    @handler.register(gnes_pb2.Response.QueryResponse)
    def _handler_doc_search(self, msg: 'gnes_pb2.Message'):
        from ..indexer.base import BaseDocIndexer
        if not isinstance(self._model, BaseDocIndexer):
            raise ServiceError(
                'unsupported indexer, dont know how to use %s to handle this message' % self._model.__bases__)

        # check if chunk_indexer and doc_indexer has the same sorting order
        if msg.response.search.is_big_score_similar is not None and \
                msg.response.search.is_big_score_similar != self._model.is_big_score_similar:
            raise ServiceError(
                'is_big_score_similar is inconsistent. last topk-list: is_big_score_similar=%s, but '
                'this indexer: is_big_score_similar=%s' % (
                    msg.response.search.is_big_score_similar, self._model.is_big_score_similar))

        # assume the doc search will change the whatever sort order the message has
        msg.response.search.is_sorted = False
        results = self._model.query_and_score(msg.response.search.topk_results)
        self._put_result_into_message(results, msg)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import time
import threading
import queue
from concurrent import futures

from .base import GrpcClient, ResponseHandler


class SyncClient(GrpcClient):
    handler = ResponseHandler(GrpcClient.handler)

    def __init__(self, args):
        super().__init__(args)
        self._pool = futures.ThreadPoolExecutor(
            max_workers=self.args.max_concurrency)

    def send_request(self, request):
        # Send requests in seperate threads to support multiple outstanding rpcs
        self._pool.submit(self.call, request)

    def close(self):
        self._pool.shutdown(wait=True)
        super().close()


class StreamingClient(GrpcClient):
    handler = ResponseHandler(GrpcClient.handler)

    def __init__(self, args):
        super().__init__(args)

        self._request_queue = queue.Queue(maxsize=1000)
        self._is_streaming = threading.Event()

        self._dispatch_thread = threading.Thread(target=self._start)
        self._dispatch_thread.setDaemon(True)

    def send_request(self, request):
        self._request_queue.put(request, block=True)

        # create a new streaming call
        if not self._is_streaming.is_set():
            self._dispatch_thread.start()

    def _start(self):
        self._is_streaming.set()
        self.stream_call(self._request_generator())
        self._is_streaming.clear()

    def _request_generator(self):
        while self._is_streaming.is_set():
            try:
                request = self._request_queue.get(block=True, timeout=5.0)
                if request is None:
                    break
                yield request
            except queue.Empty:
                continue
            except Exception as e:
                print('exception: %s' % str(e))
                break

    @handler.register(NotImplementedError)
    def _handler_default(self, resp: 'gnes_pb2.Response'):
        raise NotImplementedError

    def close(self):
        self._is_streaming.clear()
        self._dispatch_thread.join()
        super().close()



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the 'License');
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an 'AS IS' BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import subprocess as sp


def kwargs_to_cmd_args(kwargs):
    args = []
    for k, v in kwargs.items():
        args.append('-%s' % k)
        if v is not None:
            args.append('%s' % str(v))
    return args


def _check_input(input_fn: str, input_data: bytes):
    capture_stdin = (input_fn == 'pipe:')
    if capture_stdin and input_data is None:
        raise ValueError(
            "the buffered video data for stdin should not be empty")


def run_command_async(cmd_args,
                      pipe_stdin=True,
                      pipe_stdout=False,
                      pipe_stderr=False,
                      quiet=False):
    stdin_stream = sp.PIPE if pipe_stdin else None
    stdout_stream = sp.PIPE if pipe_stdout or quiet else None
    stderr_stream = sp.PIPE if pipe_stderr or quiet else None

    return sp.Popen(
        cmd_args,
        stdin=stdin_stream,
        stdout=stdout_stream,
        stderr=stderr_stream,
        close_fds=True)


def wait(process):
    while True:
        output = process.stdout.readline()
        if output == '' and process.poll() is not None:
            break
        if output:
            print(output.strip())
    rc = process.poll()
    return (output, rc)


def run_command(cmd_args,
                input=None,
                pipe_stdin=True,
                pipe_stdout=False,
                pipe_stderr=False,
                quiet=False):
    with run_command_async(
            cmd_args,
            pipe_stdin=pipe_stdin,
            pipe_stdout=pipe_stdout,
            pipe_stderr=pipe_stderr,
            quiet=quiet) as proc:
        stdout, stderr = proc.communicate(input)
        retcode = proc.poll()

        if retcode:
            raise Exception('ffmpeg error: %s' % stderr)

        if proc.stdout is not None:
            proc.stdout.close()
        if proc.stderr is not None:
            proc.stderr.close()

        return stdout, stderr



import os
import unittest
import numpy as np

from gnes.encoder.base import BaseEncoder


class TestPCAEncoder(unittest.TestCase):

    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.dump_path = os.path.join(dirname, 'pca_encoder.bin')
        self.yaml_path = os.path.join(dirname, 'yaml', 'pca.yml')
        self.test_numeric = np.random.randint(0, 255, (1000, 1024)).astype('float32')

    def test_encoding(self):
        self.encoder = BaseEncoder.load_yaml(self.yaml_path)
        # train before encode to create pca_components
        self.encoder.train(self.test_numeric)
        vec = self.encoder.encode(self.test_numeric)
        self.assertEqual(vec.shape, (1000, 300))
        # dump after train with valied pca_components
        self.encoder.dump(self.dump_path)
        encoder2 = BaseEncoder.load(self.dump_path)
        vec = encoder2.encode(self.test_numeric)
        self.assertEqual(vec.shape, (1000, 300))

    def tearDown(self):
        if os.path.exists(self.dump_path):
            os.remove(self.dump_path)



import time
import unittest

from gnes.client.cli import ProgressBar


class TestProgessbar(unittest.TestCase):
    def setUp(self):
        self.bar_len = 20

    def test_progressbar5(self):
        # should be 5 line
        with ProgressBar(task_name='test', bar_len=self.bar_len) as pb:
            for j in range(5 * self.bar_len):
                pb.update()

    def test_progressbar1(self):
        # should be single line
        with ProgressBar(task_name='test', bar_len=self.bar_len) as pb:
            for j in range(self.bar_len):
                pb.update()
                time.sleep(.1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import os
from typing import List, Tuple, Any

import numpy as np

from .cython import IndexCore
from ..helper import ListKeyIndexer
from ...base import BaseChunkIndexer as BCI


class HBIndexer(BCI):

    def __init__(self,
                 num_clusters: int = 100,
                 num_bytes: int = 8,
                 n_idx: int = 1,
                 data_path: str = './hbindexer_data',
                 *args,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.n_bytes = num_bytes
        self.n_clusters = num_clusters
        self.n_idx = n_idx
        self.data_path = data_path
        if self.n_idx <= 0:
            raise ValueError('There should be at least 1 clustering slot')
        self.helper_indexer = ListKeyIndexer()

    def post_init(self):
        self.hbindexer = IndexCore(self.n_clusters, self.n_bytes, self.n_idx)
        try:
            if not os.path.exists(self.data_path):
                raise FileNotFoundError('"data_path" is not exist')
            if os.path.isdir(self.data_path):
                raise IsADirectoryError('"data_path" must be a file path, not a directory')
            self.hbindexer.load(self.data_path)
        except (FileNotFoundError, IsADirectoryError):
            self.logger.warning('fail to load model from %s, will create an empty one' % self.data_path)

    @BCI.update_helper_indexer
    def add(self, keys: List[Tuple[int, Any]], vectors: np.ndarray, weights: List[float], *args, **kwargs):
        if len(vectors) != len(keys):
            raise ValueError("vectors length should be equal to doc_ids")

        if vectors.dtype != np.uint32:
            raise ValueError("vectors should be ndarray of uint32")

        n = len(keys)
        keys, offsets = zip(*keys)
        keys = np.array(keys, dtype=np.uint32).tobytes()
        offsets = np.array(offsets, dtype=np.uint16).tobytes()
        weights = self.float2uint_weight(weights).tobytes()
        clusters = vectors[:, :self.n_idx].tobytes()
        vectors = vectors[:, self.n_idx:].astype(np.uint8).tobytes()
        self.hbindexer.index_trie(vectors, clusters, keys, offsets, weights, n)

    @staticmethod
    def float2uint_weight(weights: List[float], norm: int = 2 ** 16 - 1):
        weights = norm * np.array(weights)
        return np.array(weights, dtype=np.uint16)

    @staticmethod
    def uint2float_weight(weight: int, norm: int = 2 ** 16 - 1):
        return weight / norm

    def query(self,
              vectors: np.ndarray,
              top_k: int,
              *args,
              **kwargs) -> List[List[Tuple]]:

        if vectors.dtype != np.uint32:
            raise ValueError("vectors should be ndarray of uint32")

        # num_rows = int(len(keys) / self.num_bytes)
        n = vectors.shape[0]
        clusters = vectors[:, :self.n_idx].tobytes()
        vectors = vectors[:, self.n_idx:].astype(np.uint8).tobytes()

        result = [{} for _ in range(n)]

        doc_ids, offsets, weights, dists, q_idx = self.hbindexer.query(
            vectors, clusters, n, top_k * self.n_idx)
        for (i, o, w, d, q) in zip(doc_ids, offsets, weights, dists, q_idx):
            result[q][(i, o, self.uint2float_weight(w))] = d

        return [list(ret.items()) for ret in result]

    def __getstate__(self):
        self.hbindexer.save(self.data_path)
        d = super().__getstate__()
        return d



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


from typing import List, Union

from .base import BaseService as BS, MessageHandler
from ..proto import gnes_pb2, array2blob, blob2array


class EncoderService(BS):
    handler = MessageHandler(BS.handler)

    def post_init(self):
        from ..encoder.base import BaseEncoder
        self._model = self.load_model(BaseEncoder)
        self.train_data = []

    def embed_chunks_in_docs(self, docs: Union[List['gnes_pb2.Document'], 'gnes_pb2.Document'],
                             do_encoding: bool = True,
                             is_input_list: bool = True):
        if not is_input_list:
            docs = [docs]

        contents = []
        chunks = []

        for d in docs:
            if not d.chunks:
                self.logger.warning('document (doc_id=%s) contains no chunks!' % d.doc_id)
                continue

            for c in d.chunks:
                if d.doc_type == gnes_pb2.Document.TEXT:
                    contents.append(c.text)
                elif d.doc_type in {gnes_pb2.Document.IMAGE, gnes_pb2.Document.VIDEO}:
                    contents.append(blob2array(c.blob))
                else:
                    self.logger.warning(
                        'chunk content is in type: %s, dont kow how to handle that, ignored' % c.WhichOneof('content'))
                chunks.append(c)

        if do_encoding and contents:
            try:
                embeds = self._model.encode(contents)
                if len(chunks) != embeds.shape[0]:
                    self.logger.error(
                        'mismatched %d chunks and a %s shape embedding, '
                        'the first dimension must be the same' % (len(chunks), embeds.shape))
                for idx, c in enumerate(chunks):
                    c.embedding.CopyFrom(array2blob(embeds[idx]))
            except Exception as ex:
                self.logger.error(ex, exc_info=True)
                self.logger.warning('encoder service throws an exception, '
                                    'the sequel pipeline may not work properly')

        return contents

    @handler.register(gnes_pb2.Request.IndexRequest)
    def _handler_index(self, msg: 'gnes_pb2.Message'):
        self.embed_chunks_in_docs(msg.request.index.docs)

    @handler.register(gnes_pb2.Request.TrainRequest)
    def _handler_train(self, msg: 'gnes_pb2.Message'):
        if msg.request.train.docs:
            contents = self.embed_chunks_in_docs(msg.request.train.docs, do_encoding=False)
            self.train_data.extend(contents)
            msg.response.train.status = gnes_pb2.Response.PENDING
            # raise BlockMessage
        if msg.request.train.flush:
            self._model.train(self.train_data)
            self.logger.info('%d samples is flushed for training' % len(self.train_data))
            self.is_model_changed.set()
            self.train_data.clear()
            msg.response.control.status = gnes_pb2.Response.SUCCESS

    @handler.register(gnes_pb2.Request.QueryRequest)
    def _handler_search(self, msg: 'gnes_pb2.Message'):
        self.embed_chunks_in_docs(msg.request.search.query, is_input_list=False)

    @handler.register_hook(hook_type=('pre', 'post'), only_when_verbose=True)
    def _hook_debug_msg(self, msg: 'gnes_pb2.Message', *args, **kwargs):
        from pprint import pformat

        debug_kv = {
            'envelope': lambda: msg.envelope,
            'num_docs': lambda: len(msg.request.index.docs),
            'num_chunks in doc[0]': lambda: len(msg.request.index.docs[0].chunks),
            'docs[0].chunks[0].content_type': lambda: msg.request.index.docs[0].chunks[0].WhichOneof('content'),
            'docs[0].chunks[0].weight': lambda: msg.request.index.docs[0].chunks[0].weight,
            'docs[0].chunks[0].embedding': lambda: blob2array(msg.request.index.docs[0].chunks[0].embedding),
            'docs[0].chunks[0].embedding[0]': lambda: blob2array(msg.request.index.docs[0].chunks[0].embedding)[0]
        }
        debug_info = {}
        for k, v in debug_kv.items():
            try:
                r = v()
            except Exception as ex:
                r = 'fail to get the value, reason: %s' % ex
            debug_info[k] = r
        self.logger.info(pformat(debug_info))



import os
import unittest

import grpc

from gnes.cli.parser import set_frontend_parser, _set_loadable_service_parser, set_preprocessor_parser
from gnes.proto import gnes_pb2_grpc, RequestGenerator
from gnes.service.base import ServiceManager, SocketType
from gnes.service.encoder import EncoderService
from gnes.service.frontend import FrontendService
from gnes.service.preprocessor import PreprocessorService


@unittest.SkipTest
class TestEncoder(unittest.TestCase):

    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.transformer_yml = os.path.join(self.dirname, 'contrib', 'transformer.yml')
        self.transformer_py = os.path.join(self.dirname, 'contrib', 'transformer.py')
        os.environ['GNES_CONTRIB_MODULE'] = self.transformer_py
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')

    def test_pymode(self):
        args = set_frontend_parser().parse_args([])

        p_args = set_preprocessor_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', '5531',
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_BIND),
            '--yaml_path', '!UnaryPreprocessor {parameters: {doc_type: 1}}'
        ])

        e_args = _set_loadable_service_parser().parse_args([
            '--port_in', str(p_args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', self.transformer_yml,
        ])

        with ServiceManager(EncoderService, e_args), \
             ServiceManager(PreprocessorService, p_args), \
             ServiceManager(FrontendService, args), \
             grpc.insecure_channel('%s:%d' % (args.grpc_host, args.grpc_port),
                                   options=[('grpc.max_send_message_length', 70 * 1024 * 1024),
                                            ('grpc.max_receive_message_length', 70 * 1024 * 1024)]) as channel:
            stub = gnes_pb2_grpc.GnesRPCStub(channel)
            resp = stub.Call(list(RequestGenerator.index([b'hello world', b'goodbye!'], 1))[0])
            self.assertEqual(resp.request_id, 0)



import os
import unittest

from gnes.encoder.base import BaseEncoder
from gnes.helper import PathImporter


class TestPipeTrain(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        PathImporter.add_modules(*('{0}/contrib/dummy2.py,{0}/contrib/dummy3.py'.format(self.dirname).split(',')))

    def tearDown(self):
        if os.path.exists('dummy-pipeline.bin'):
            os.remove('dummy-pipeline.bin')
        if os.path.exists('dummy-pipeline.yml'):
            os.remove('dummy-pipeline.yml')

    def test_load_yaml(self):
        p = BaseEncoder.load_yaml(os.path.join(self.dirname, 'yaml', 'pipeline-multi-encoder2.yml'))
        self.assertFalse(p.is_trained)
        self.assertRaises(RuntimeError, p.encode, 1)
        p.train(1)
        self.assertTrue(p.is_trained)
        self.assertEqual(5, p.encode(1))
        p = BaseEncoder.load_yaml(os.path.join(self.dirname, 'yaml', 'pipeline-multi-encoder2.yml'))
        self.assertRaises(RuntimeError, p.encode, 1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from .base import get_unary_score, CombinedScoreFn
import json


class WeightedDocScoreFn(CombinedScoreFn):
    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 doc: 'gnes_pb2.Document', *args, **kwargs):
        d_weight = get_unary_score(value=doc.weight,
                                   name='doc weight',
                                   doc_id=doc.doc_id)
        return super().__call__(last_score, d_weight)


class CoordDocScoreFn(CombinedScoreFn):
    """
    score = score * query_coordination
    query_coordination: #chunks recalled / #chunks in this doc
    """

    def __call__(self, last_score: 'gnes_pb2.Response.QueryResponse.ScoredResult.Score',
                 doc: 'gnes_pb2.Document',
                 *args, **kwargs):
        total_chunks = len(doc.chunks)
        recall_chunks = len(json.loads(last_score.explained)['operands'])
        query_coord = 1 if total_chunks == 0 else recall_chunks / total_chunks
        d_weight = get_unary_score(value=query_coord,
                                   name='query coordination')
        return super().__call__(last_score, d_weight)




import unittest

import numpy as np
import torch
from numpy.testing import assert_allclose

from gnes.encoder.numeric.pooling import PoolingEncoder


class TestEncoder(unittest.TestCase):
    def setUp(self):
        self.seq_data = np.random.random([5, 10])
        self.seq_embed_data = np.random.random([5, 10, 32])
        self.mask_data = np.array(self.seq_data > 0.5, np.float32)
        self.data = [
            (torch.tensor(self.seq_embed_data, dtype=torch.float32), torch.tensor(self.mask_data, dtype=torch.float32)),
            (self.seq_embed_data, self.mask_data),
            (self.seq_embed_data, self.mask_data)]

    def _test_strategy(self, strategy):
        pe_to = PoolingEncoder(strategy, 'torch')
        pe_tf = PoolingEncoder(strategy, 'tensorflow')
        pe_np = PoolingEncoder(strategy, 'numpy')
        return [pe.encode(self.data[idx]) for idx, pe in enumerate([pe_to, pe_tf, pe_np])]

    def test_all(self):
        for s in {'REDUCE_MEAN', 'REDUCE_MAX', 'REDUCE_MEAN_MAX'}:
            with self.subTest(strategy=s):
                r = self._test_strategy(s)
                for rr in r:
                    print(type(rr))
                    print(rr)
                    print('---')
                assert_allclose(r[0], r[1], rtol=1e-5)
                assert_allclose(r[1], r[2], rtol=1e-5)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ctypes
import random
from typing import List, Iterator
from typing import Optional

import numpy as np
import zmq
from termcolor import colored

from . import gnes_pb2
from ..helper import batch_iterator, default_logger

__all__ = ['RequestGenerator', 'send_message', 'recv_message', 'blob2array', 'array2blob', 'gnes_pb2', 'add_route']


class RequestGenerator:
    @staticmethod
    def index(data: Iterator[bytes], batch_size: int = 0, doc_type: int = gnes_pb2.Document.TEXT,
              doc_id_start: int = 0, request_id_start: int = 0,
              random_doc_id: bool = False,
              *args, **kwargs):

        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = request_id_start
            for raw_bytes in pi:
                d = req.index.docs.add()
                d.doc_id = doc_id_start if not random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
                d.raw_bytes = raw_bytes
                d.weight = 1.0
                d.doc_type = doc_type
                doc_id_start += 1
            yield req
            request_id_start += 1

    @staticmethod
    def train(data: Iterator[bytes], batch_size: int = 0, doc_type: int = gnes_pb2.Document.TEXT,
              doc_id_start: int = 0, request_id_start: int = 0,
              random_doc_id: bool = False,
              *args, **kwargs):
        for pi in batch_iterator(data, batch_size):
            req = gnes_pb2.Request()
            req.request_id = request_id_start
            for raw_bytes in pi:
                d = req.train.docs.add()
                d.doc_id = doc_id_start if not random_doc_id else random.randint(0, ctypes.c_uint(-1).value)
                d.raw_bytes = raw_bytes
                d.doc_type = doc_type
                if not random_doc_id:
                    doc_id_start += 1
            yield req
            request_id_start += 1
        req = gnes_pb2.Request()
        req.request_id = request_id_start
        req.train.flush = True
        yield req
        request_id_start += 1

    @staticmethod
    def query(query: bytes, top_k: int, request_id_start: int = 0, doc_type: int = gnes_pb2.Document.TEXT, *args,
              **kwargs):
        if top_k <= 0:
            raise ValueError('"top_k: %d" is not a valid number' % top_k)

        req = gnes_pb2.Request()
        req.request_id = request_id_start
        req.search.query.raw_bytes = query
        req.search.query.doc_type = doc_type
        req.search.top_k = top_k
        yield req


def blob2array(blob: 'gnes_pb2.NdArray') -> np.ndarray:
    """
    Convert a blob proto to an array.
    """
    return np.frombuffer(blob.data, dtype=blob.dtype).reshape(blob.shape)


def array2blob(x: np.ndarray) -> 'gnes_pb2.NdArray':
    """Converts a N-dimensional array to blob proto.
    """
    blob = gnes_pb2.NdArray()
    blob.data = x.tobytes()
    blob.shape.extend(list(x.shape))
    blob.dtype = x.dtype.name
    return blob


def router2str(m: 'gnes_pb2.Message') -> str:
    route_str = [r.service for r in m.envelope.routes]
    return colored('▸', 'green').join(route_str)


def add_route(evlp: 'gnes_pb2.Envelope', name: str, identity: str):
    r = evlp.routes.add()
    r.service = name
    r.start_time.GetCurrentTime()
    r.service_identity = identity


def merge_routes(msg: 'gnes_pb2.Message', prev_msgs: List['gnes_pb2.Message']):
    # take unique routes by service identity
    routes = {r.service_identity: r for m in prev_msgs for r in m.envelope.routes}
    msg.envelope.ClearField('routes')
    msg.envelope.routes.extend(sorted(routes.values(), key=lambda x: (x.start_time.seconds, x.start_time.nanos)))


def send_message(sock: 'zmq.Socket', msg: 'gnes_pb2.Message', timeout: int = -1) -> None:
    try:
        if timeout > 0:
            sock.setsockopt(zmq.SNDTIMEO, timeout)
        else:
            sock.setsockopt(zmq.SNDTIMEO, -1)

        sock.send_multipart([msg.envelope.client_id.encode(), msg.SerializeToString()])
    except zmq.error.Again:
        raise TimeoutError(
            'cannot send message to sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    except Exception as ex:
        raise ex
    finally:
        sock.setsockopt(zmq.SNDTIMEO, -1)


def recv_message(sock: 'zmq.Socket', timeout: int = -1, check_version: bool = False) -> Optional['gnes_pb2.Message']:
    response = []
    try:
        if timeout > 0:
            sock.setsockopt(zmq.RCVTIMEO, timeout)
        else:
            sock.setsockopt(zmq.RCVTIMEO, -1)

        _, msg_data = sock.recv_multipart()
        msg = gnes_pb2.Message()
        msg.ParseFromString(msg_data)

        if check_version and msg.envelope:
            from .. import __version__, __proto_version__
            if hasattr(msg.envelope, 'gnes_version'):
                if not msg.envelope.gnes_version:
                    # only happen in unittest
                    default_logger.warning('incoming message contains empty "gnes_version", '
                                           'you may ignore it in debug/unittest mode. '
                                           'otherwise please check if frontend service set correct version')
                elif __version__ != msg.envelope.gnes_version:
                    raise AttributeError('mismatched GNES version! '
                                         'incoming message has GNES version %s, whereas local GNES version %s' % (
                                             msg.envelope.gnes_version, __version__))
            if hasattr(msg.envelope, 'proto_version'):
                if not msg.envelope.proto_version:
                    # only happen in unittest
                    default_logger.warning('incoming message contains empty "proto_version", '
                                           'you may ignore it in debug/unittest mode. '
                                           'otherwise please check if frontend service set correct version')
                elif __proto_version__ != msg.envelope.proto_version:
                    raise AttributeError('mismatched protobuf version! '
                                         'incoming message has protobuf version %s, whereas local protobuf version %s' % (
                                             msg.envelope.proto_version, __proto_version__))
            if not hasattr(msg.envelope, 'proto_version') and not hasattr(msg.envelope, 'gnes_version'):
                raise AttributeError('version_check=True locally, '
                                     'but incoming message contains no version info in its envelope. '
                                     'the message is probably sent from a very outdated GNES version')
        return msg

    except ValueError:
        raise ValueError('received a wrongly-formatted request (expected 4 frames, got %d)' % len(response))
    except zmq.error.Again:
        raise TimeoutError(
            'no response from sock %s after timeout=%dms, please check the following:'
            'is the server still online? is the network broken? are "port" correct? ' % (
                sock, timeout))
    except Exception as ex:
        raise ex
    finally:
        sock.setsockopt(zmq.RCVTIMEO, -1)



#  Tencent is pleased to support the open source community by making GNES available.
#
#  Copyright (C) 2019 THL A29 Limited, a Tencent company. All rights reserved.
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.


import sys
import time
import zipfile
from typing import Generator

from termcolor import colored

from .base import GrpcClient
from ..proto import RequestGenerator, gnes_pb2


class CLIClient(GrpcClient):
    def __init__(self, args):
        super().__init__(args)
        getattr(self, self.args.mode)()
        self.close()

    def train(self):
        with ProgressBar(task_name=self.args.mode) as p_bar:
            for _ in self._stub.StreamCall(RequestGenerator.train(self.bytes_generator,
                                                                  doc_id_start=self.args.start_doc_id,
                                                                  batch_size=self.args.batch_size)):
                p_bar.update()

    def index(self):
        with ProgressBar(task_name=self.args.mode) as p_bar:
            for _ in self._stub.StreamCall(RequestGenerator.index(self.bytes_generator,
                                                                  doc_id_start=self.args.start_doc_id,
                                                                  batch_size=self.args.batch_size)):
                p_bar.update()

    def query(self):
        for idx, q in enumerate(self.bytes_generator):
            for req in RequestGenerator.query(q, request_id_start=idx, top_k=self.args.top_k):
                resp = self._stub.Call(req)
                self.query_callback(req, resp)

    def query_callback(self, req: 'gnes_pb2.Request', resp: 'gnes_pb2.Response'):
        """
        callback after get the query result
        override this method to customize query behavior
        :param resp: response
        :param req: query
        :return:
        """
        print(req)
        print(resp)

    @property
    def bytes_generator(self) -> Generator[bytes, None, None]:
        if self.args.txt_file:
            all_bytes = (v.encode() for v in self.args.txt_file)
        elif self.args.image_zip_file:
            zipfile_ = zipfile.ZipFile(self.args.image_zip_file)
            all_bytes = (zipfile_.open(v).read() for v in zipfile_.namelist())
        elif self.args.video_zip_file:
            zipfile_ = zipfile.ZipFile(self.args.video_zip_file)
            all_bytes = (zipfile_.open(v).read() for v in zipfile_.namelist())
        else:
            raise AttributeError('--txt_file, --image_zip_file, --video_zip_file one must be given')

        return all_bytes


class ProgressBar:
    def __init__(self, bar_len: int = 20, task_name: str = ''):
        self.bar_len = bar_len
        self.task_name = task_name

    def update(self):
        sys.stdout.write('\r')
        elapsed = time.perf_counter() - self.start_time
        elapsed_str = colored('elapsed', 'yellow')
        speed_str = colored('speed', 'yellow')
        self.num_bars += 1
        if self.num_bars > self.bar_len:
            self.num_bars -= self.bar_len
            sys.stdout.write('\n')
        sys.stdout.write(
            '{:>10} [{:<{}}]  {:>8}: {:3.1f}s   {:>8}: {:3.1f} batch/s'.format(
                colored(self.task_name, 'cyan'),
                colored('=' * self.num_bars, 'green'),
                self.bar_len + 9,
                elapsed_str,
                elapsed,
                speed_str,
                self.num_bars / elapsed,
            ))
        sys.stdout.flush()

    def __enter__(self):
        self.start_time = time.perf_counter()
        self.num_bars = -1
        sys.stdout.write('\n')
        self.update()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        sys.stdout.write('\t%s\n' % colored('done!', 'green'))



import os
import unittest

from gnes.cli.parser import set_frontend_parser, set_router_parser, set_client_cli_parser
from gnes.client.cli import CLIClient
from gnes.service.base import SocketType
from gnes.service.frontend import FrontendService
from gnes.service.router import RouterService


class TestCLI(unittest.TestCase):
    def setUp(self):
        self.dirname = os.path.dirname(__file__)
        self.test_file = os.path.join(self.dirname, 'sonnets_small.txt')
        self.train_args = set_client_cli_parser().parse_args([
            '--mode', 'train',
            '--txt_file', self.test_file
        ])
        self.index_args = set_client_cli_parser().parse_args([
            '--mode', 'index',
            '--txt_file', self.test_file
        ])
        self.query_args = set_client_cli_parser().parse_args([
            '--mode', 'query',
            '--txt_file', self.test_file
        ])
        os.unsetenv('http_proxy')
        os.unsetenv('https_proxy')

    def test_cli(self):
        args = set_frontend_parser().parse_args([])

        p_args = set_router_parser().parse_args([
            '--port_in', str(args.port_out),
            '--port_out', str(args.port_in),
            '--socket_in', str(SocketType.PULL_CONNECT),
            '--socket_out', str(SocketType.PUSH_CONNECT),
            '--yaml_path', 'BaseRouter'
        ])

        with RouterService(p_args), FrontendService(args):
            CLIClient(self.train_args)
            CLIClient(self.index_args)
            CLIClient(self.query_args)



import os
import unittest
import numpy as np

from gnes.encoder.base import BaseEncoder


class TestYT8MEncoder(unittest.TestCase):
    @unittest.skip
    def setUp(self):
        dirname = os.path.dirname(__file__)
        self.test_video = [np.random.rand(10, 1152).astype(np.uint8),
                         np.random.rand(15, 1152).astype(np.uint8),
                         np.random.rand(20, 1152).astype(np.uint8)]
        self.yt8m_yaml = os.path.join(dirname, 'yaml', 'yt8m_encoder.yml')

    @unittest.skip
    def test_yt8m_encoding(self):
        self.encoder = BaseEncoder.load_yaml(self.yt8m_yaml)
        vec = self.encoder.encode(self.test_video)

        self.assertEqual(vec.shape[0], len(self.test_video))
        self.assertEqual(vec.shape[1], 19310)